@@ -25,18 +25,24 @@ import time
 
 from keystoneauth1.access import service_catalog as keystone_sc
 from keystoneauth1 import exceptions as keystone_exc
+from keystoneauth1 import identity as ksa_identity
+from keystoneauth1 import session as ksa_session
+from keystoneauth1 import token_endpoint as ksa_token_endpoint
 from oslo_concurrency import processutils
 from oslo_config import cfg
 from oslo_utils import units
 
 from glance_store import capabilities
+from glance_store.common import attachment_state_manager
+from glance_store.common import cinder_utils
 from glance_store.common import utils
 import glance_store.driver
 from glance_store import exceptions
-from glance_store.i18n import _, _LE, _LI
+from glance_store.i18n import _, _LE, _LI, _LW
 import glance_store.location
 
 try:
+    from cinderclient import api_versions
     from cinderclient import exceptions as cinder_exception
     from cinderclient.v3 import client as cinderclient
     from os_brick.initiator import connector
@@ -53,7 +59,7 @@ LOG = logging.getLogger(__name__)
 
 _CINDER_OPTS = [
     cfg.StrOpt('cinder_catalog_info',
-               default='volumev2::publicURL',
+               default='volumev3::publicURL',
                help="""
 Information to match when looking for cinder in the service catalog.
 
@@ -79,6 +85,8 @@ Related options:
     * cinder_store_user_name
     * cinder_store_project_name
     * cinder_store_password
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
 
 """),
     cfg.StrOpt('cinder_endpoint_template',
@@ -104,6 +112,8 @@ Related options:
     * cinder_store_user_name
     * cinder_store_project_name
     * cinder_store_password
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
     * cinder_catalog_info
 
 """),
@@ -215,6 +225,8 @@ Related options:
     * cinder_store_user_name
     * cinder_store_password
     * cinder_store_project_name
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
 
 """),
     cfg.StrOpt('cinder_store_user_name',
@@ -222,8 +234,9 @@ Related options:
                help="""
 User name to authenticate against cinder.
 
-This must be used with all the following related options. If any of these are
-not specified, the user of the current context is used.
+This must be used with all the following non-domain-related options.
+If any of these are not specified (except domain-related options),
+the user of the current context is used.
 
 Possible values:
     * A valid user name
@@ -232,14 +245,33 @@ Related options:
     * cinder_store_auth_address
     * cinder_store_password
     * cinder_store_project_name
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
+
+"""),
+    cfg.StrOpt('cinder_store_user_domain_name',
+               default='Default',
+               help="""
+Domain of the user to authenticate against cinder.
+
+Possible values:
+    * A valid domain name for the user specified by ``cinder_store_user_name``
+
+Related options:
+    * cinder_store_auth_address
+    * cinder_store_password
+    * cinder_store_project_name
+    * cinder_store_project_domain_name
+    * cinder_store_user_name
 
 """),
     cfg.StrOpt('cinder_store_password', secret=True,
                help="""
 Password for the user authenticating against cinder.
 
-This must be used with all the following related options. If any of these are
-not specified, the user of the current context is used.
+This must be used with all the following related options.
+If any of these are not specified (except domain-related options),
+the user of the current context is used.
 
 Possible values:
     * A valid password for the user specified by ``cinder_store_user_name``
@@ -248,6 +280,8 @@ Related options:
     * cinder_store_auth_address
     * cinder_store_user_name
     * cinder_store_project_name
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
 
 """),
     cfg.StrOpt('cinder_store_project_name',
@@ -258,8 +292,9 @@ Project name where the image volume is stored in cinder.
 If this configuration option is not set, the project in current context is
 used.
 
-This must be used with all the following related options. If any of these are
-not specified, the project of the current context is used.
+This must be used with all the following related options.
+If any of these are not specified (except domain-related options),
+the user of the current context is used.
 
 Possible values:
     * A valid project name
@@ -268,6 +303,25 @@ Related options:
     * ``cinder_store_auth_address``
     * ``cinder_store_user_name``
     * ``cinder_store_password``
+    * ``cinder_store_project_domain_name``
+    * ``cinder_store_user_domain_name``
+
+"""),
+    cfg.StrOpt('cinder_store_project_domain_name',
+               default='Default',
+               help="""
+Domain of the project where the image volume is stored in cinder.
+
+Possible values:
+    * A valid domain name of the project specified by
+      ``cinder_store_project_name``
+
+Related options:
+    * ``cinder_store_auth_address``
+    * ``cinder_store_user_name``
+    * ``cinder_store_password``
+    * ``cinder_store_project_domain_name``
+    * ``cinder_store_user_domain_name``
 
 """),
     cfg.StrOpt('rootwrap_config',
@@ -328,7 +382,7 @@ Related options:
     cfg.BoolOpt('cinder_use_multipath',
                 default=False,
                 help="""
-Flag to identify mutipath is supported or not in the deployment.
+Flag to identify multipath is supported or not in the deployment.
 
 Set it to False if multipath is not supported.
 
@@ -350,6 +404,34 @@ Possible values:
 """),
 ]
 
+CINDER_SESSION = None
+
+
+def _reset_cinder_session():
+    global CINDER_SESSION
+    CINDER_SESSION = None
+
+
+def get_cinder_session(conf):
+    global CINDER_SESSION
+    if not CINDER_SESSION:
+        auth = ksa_identity.V3Password(
+            password=conf.cinder_store_password,
+            username=conf.cinder_store_user_name,
+            user_domain_name=conf.cinder_store_user_domain_name,
+            project_name=conf.cinder_store_project_name,
+            project_domain_name=conf.cinder_store_project_domain_name,
+            auth_url=conf.cinder_store_auth_address
+        )
+        if conf.cinder_api_insecure:
+            verify = False
+        elif conf.cinder_ca_certificates_file:
+            verify = conf.cinder_ca_certificates_file
+        else:
+            verify = True
+        CINDER_SESSION = ksa_session.Session(auth=auth, verify=verify)
+    return CINDER_SESSION
+
 
 class StoreLocation(glance_store.location.StoreLocation):
 
@@ -397,6 +479,7 @@ class Store(glance_store.driver.Store):
             self.store_conf = getattr(self.conf, self.backend_group)
         else:
             self.store_conf = self.conf.glance_store
+        self.volume_api = cinder_utils.API()
 
     def _set_url_prefix(self):
         self._url_prefix = "cinder://"
@@ -405,38 +488,26 @@ class Store(glance_store.driver.Store):
 
     def configure_add(self):
         """
-        Configure the Store to use the stored configuration options
-        Any store that needs special configuration should implement
-        this method. If the store was not able to successfully configure
-        itself, it should raise `exceptions.BadStoreConfiguration`
-        :raises: `exceptions.BadStoreConfiguration` if multiple stores are
-                 defined and particular store wasn't able to configure
-                 successfully
-        :raises: `exceptions.BackendException` if single store is defined and
-                 it wasn't able to configure successfully
+        Check to verify if the volume types configured for the cinder store
+        exist in deployment and if not, log a warning.
         """
-        if self.backend_group:
-            cinder_volume_type = self.store_conf.cinder_volume_type
-            if cinder_volume_type:
-                # NOTE: `cinder_volume_type` is configured, check
-                # configured volume_type is available in cinder or not
-                cinder_client = self.get_cinderclient()
-                try:
-                    # We don't even need the volume type object, as long
-                    # as this returns clean, we know the name is good.
-                    cinder_client.volume_types.find(name=cinder_volume_type)
-                    # No need to worry NoUniqueMatch as volume type name is
-                    # unique
-                except cinder_exception.NotFound:
-                    reason = _("Invalid `cinder_volume_type %s`"
-                               % cinder_volume_type)
-                    if len(self.conf.enabled_backends) > 1:
-                        LOG.error(reason)
-                        raise exceptions.BadStoreConfiguration(
-                            store_name=self.backend_group, reason=reason)
-                    else:
-                        LOG.critical(reason)
-                        raise exceptions.BackendException(reason)
+        cinder_volume_type = self.store_conf.cinder_volume_type
+        if cinder_volume_type:
+            # NOTE: `cinder_volume_type` is configured, check
+            # configured volume_type is available in cinder or not
+            cinder_client = self.get_cinderclient()
+            try:
+                # We don't even need the volume type object, as long
+                # as this returns clean, we know the name is good.
+                cinder_client.volume_types.find(name=cinder_volume_type)
+                # No need to worry about a NoUniqueMatch as volume type name
+                # is unique
+            except cinder_exception.NotFound:
+                reason = (_LW("Invalid `cinder_volume_type %s`"
+                              % cinder_volume_type))
+                LOG.warning(reason)
+            except cinder_exception.ClientException:
+                pass
 
     def is_image_associated_with_store(self, context, volume_id):
         """
@@ -461,7 +532,7 @@ class Store(glance_store.driver.Store):
                     return True
         except Exception:
             # Glance calls this method to update legacy images URL
-            # If an exception occours due to image/volume is non-existent or
+            # If an exception occurs due to image/volume is non-existent or
             # any other reason, we return False (i.e. the image location URL
             # won't be updated) and it is glance's responsibility to handle
             # the case when the image failed to update
@@ -478,7 +549,8 @@ class Store(glance_store.driver.Store):
                     for key in ['user_name', 'password',
                                 'project_name', 'auth_address']])
 
-    def get_cinderclient(self, context=None, legacy_update=False):
+    def get_cinderclient(self, context=None, legacy_update=False,
+                         version='3.0'):
         # NOTE: For legacy image update from single store to multiple
         # stores we need to use admin context rather than user provided
         # credentials
@@ -488,15 +560,18 @@ class Store(glance_store.driver.Store):
         else:
             user_overriden = self.is_user_overriden()
 
+        session = get_cinder_session(self.store_conf)
+
         if user_overriden:
             username = self.store_conf.cinder_store_user_name
-            password = self.store_conf.cinder_store_password
-            project = self.store_conf.cinder_store_project_name
             url = self.store_conf.cinder_store_auth_address
+            # use auth that is already in the session
+            auth = None
         else:
-            username = context.user
-            password = context.auth_token
-            project = context.tenant
+            username = context.user_id
+            project = context.project_id
+            # noauth extracts user_id:project_id from auth_token
+            token = context.auth_token or '%s:%s' % (username, project)
 
             if self.store_conf.cinder_endpoint_template:
                 template = self.store_conf.cinder_endpoint_template
@@ -516,23 +591,19 @@ class Store(glance_store.driver.Store):
                     reason = _("Failed to find Cinder from a service catalog.")
                     raise exceptions.BadStoreConfiguration(store_name="cinder",
                                                            reason=reason)
+            auth = ksa_token_endpoint.Token(endpoint=url, token=token)
 
+        api_version = api_versions.APIVersion(version)
         c = cinderclient.Client(
-            username, password, project, auth_url=url,
+            session=session, auth=auth,
             region_name=self.store_conf.cinder_os_region_name,
-            insecure=self.store_conf.cinder_api_insecure,
             retries=self.store_conf.cinder_http_retries,
-            cacert=self.store_conf.cinder_ca_certificates_file)
+            api_version=api_version)
 
         LOG.debug(
             'Cinderclient connection created for user %(user)s using URL: '
             '%(url)s.', {'user': username, 'url': url})
 
-        # noauth extracts user_id:project_id from auth_token
-        if not user_overriden:
-            c.client.auth_token = context.auth_token or '%s:%s' % (username,
-                                                                   project)
-        c.client.management_url = url
         return c
 
     @contextlib.contextmanager
@@ -570,6 +641,37 @@ class Store(glance_store.driver.Store):
             raise exceptions.BadStoreConfiguration(store_name="cinder",
                                                    reason=reason)
 
+    @staticmethod
+    def _get_device_size(device_file):
+        # The seek position is corrected after every extend operation
+        # with the bytes written (which is after this wait call) so we
+        # don't need to worry about setting it back to original position
+        device_file.seek(0, os.SEEK_END)
+        # There are other ways to determine the file size like os.stat
+        # or os.path.getsize but it requires file name attribute which
+        # we don't have for the RBD file wrapper RBDVolumeIOWrapper
+        device_size = device_file.tell()
+        device_size = int(math.ceil(float(device_size) / units.Gi))
+        return device_size
+
+    @staticmethod
+    def _wait_resize_device(volume, device_file):
+        timeout = 20
+        max_recheck_wait = 10
+        tries = 0
+        elapsed = 0
+        while Store._get_device_size(device_file) < volume.size:
+            wait = min(0.5 * 2 ** tries, max_recheck_wait)
+            time.sleep(wait)
+            tries += 1
+            elapsed += wait
+            if elapsed >= timeout:
+                msg = (_('Timeout while waiting while volume %(volume_id)s '
+                         'to resize the device in %(tries)s tries.')
+                       % {'volume_id': volume.id, 'tries': tries})
+                LOG.error(msg)
+                raise exceptions.BackendException(msg)
+
     def _wait_volume_status(self, volume, status_transition, status_expected):
         max_recheck_wait = 15
         timeout = self.store_conf.cinder_state_transition_timeout
@@ -614,6 +716,12 @@ class Store(glance_store.driver.Store):
         """
         return os.path.join(mount_point_base, self.get_hash_str(share))
 
+    def _get_host_ip(self, host):
+        try:
+            return socket.getaddrinfo(host, None, socket.AF_INET6)[0][4][0]
+        except socket.gaierror:
+            return socket.getaddrinfo(host, None, socket.AF_INET)[0][4][0]
+
     @contextlib.contextmanager
     def _open_cinder_volume(self, client, volume, mode):
         attach_mode = 'rw' if mode == 'wb' else 'ro'
@@ -621,55 +729,69 @@ class Store(glance_store.driver.Store):
         root_helper = self.get_root_helper()
         priv_context.init(root_helper=shlex.split(root_helper))
         host = socket.gethostname()
+        my_ip = self._get_host_ip(host)
         use_multipath = self.store_conf.cinder_use_multipath
         enforce_multipath = self.store_conf.cinder_enforce_multipath
         mount_point_base = self.store_conf.cinder_mount_point_base
+        volume_id = volume.id
 
-        properties = connector.get_connector_properties(
-            root_helper, host, use_multipath, enforce_multipath)
+        connector_prop = connector.get_connector_properties(
+            root_helper, my_ip, use_multipath, enforce_multipath, host=host)
+
+        if volume.multiattach:
+            attachment = attachment_state_manager.attach(client, volume_id,
+                                                         host,
+                                                         mode=attach_mode)
+        else:
+            attachment = self.volume_api.attachment_create(client, volume_id,
+                                                           mode=attach_mode)
+        attachment = self.volume_api.attachment_update(
+            client, attachment['id'], connector_prop,
+            mountpoint='glance_store')
+        volume = volume.manager.get(volume_id)
+        connection_info = attachment.connection_info
 
         try:
-            volume.reserve(volume)
-        except cinder_exception.ClientException as e:
-            msg = (_('Failed to reserve volume %(volume_id)s: %(error)s')
-                   % {'volume_id': volume.id, 'error': e})
-            LOG.error(msg)
-            raise exceptions.BackendException(msg)
-
-        try:
-            connection_info = volume.initialize_connection(volume, properties)
             conn = connector.InitiatorConnector.factory(
                 connection_info['driver_volume_type'], root_helper,
                 conn=connection_info, use_multipath=use_multipath)
             if connection_info['driver_volume_type'] == 'nfs':
-                if volume.encrypted:
-                    volume.unreserve(volume)
-                    volume.delete()
-                    msg = (_('Encrypted volume creation for cinder nfs is not '
-                             'supported from glance_store. Failed to create '
-                             'volume %(volume_id)s')
-                           % {'volume_id': volume.id})
+                # The format info of nfs volumes is exposed via attachment_get
+                # API hence it is not available in the connection info of
+                # attachment object received from attachment_update and we
+                # need to do this call
+                vol_attachment = self.volume_api.attachment_get(
+                    client, attachment.id)
+                if (volume.encrypted or
+                        vol_attachment.connection_info['format'] == 'qcow2'):
+                    issue_type = 'Encrypted' if volume.encrypted else 'qcow2'
+                    msg = (_('%(issue_type)s volume creation for cinder nfs '
+                             'is not supported from glance_store. Failed to '
+                             'create volume %(volume_id)s')
+                           % {'issue_type': issue_type,
+                              'volume_id': volume_id})
                     LOG.error(msg)
                     raise exceptions.BackendException(msg)
 
-                @utils.synchronized(connection_info['data']['export'])
+                @utils.synchronized(connection_info['export'])
                 def connect_volume_nfs():
-                    data = connection_info['data']
-                    export = data['export']
-                    vol_name = data['name']
+                    export = connection_info['export']
+                    vol_name = connection_info['name']
                     mountpoint = self._get_mount_path(
                         export,
                         os.path.join(mount_point_base, 'nfs'))
-                    options = data['options']
+                    options = connection_info['options']
                     self.mount.mount(
                         'nfs', export, vol_name, mountpoint, host,
                         root_helper, options)
                     return {'path': os.path.join(mountpoint, vol_name)}
                 device = connect_volume_nfs()
             else:
-                device = conn.connect_volume(connection_info['data'])
-            volume.attach(None, 'glance_store', attach_mode, host_name=host)
-            volume = self._wait_volume_status(volume, 'attaching', 'in-use')
+                device = conn.connect_volume(connection_info)
+
+            # Complete the attachment (marking the volume "in-use") after
+            # the connection with os-brick is complete
+            self.volume_api.attachment_complete(client, attachment.id)
             if (connection_info['driver_volume_type'] == 'rbd' and
                not conn.do_local_attach):
                 yield device['path']
@@ -682,38 +804,32 @@ class Store(glance_store.driver.Store):
                               '%(volume_id)s.'), {'volume_id': volume.id})
             raise
         finally:
-            if volume.status == 'in-use':
-                volume.begin_detaching(volume)
-            elif volume.status == 'attaching':
-                volume.unreserve(volume)
-
             if device:
                 try:
                     if connection_info['driver_volume_type'] == 'nfs':
-                        @utils.synchronized(connection_info['data']['export'])
+                        @utils.synchronized(connection_info['export'])
                         def disconnect_volume_nfs():
                             path, vol_name = device['path'].rsplit('/', 1)
                             self.mount.umount(vol_name, path, host,
                                               root_helper)
                         disconnect_volume_nfs()
                     else:
-                        conn.disconnect_volume(connection_info['data'], device)
+                        if volume.multiattach:
+                            attachment_state_manager.detach(
+                                client, attachment.id, volume_id, host, conn,
+                                connection_info, device)
+                        else:
+                            # Bug #2004555: use force so there aren't any
+                            # leftovers
+                            conn.disconnect_volume(connection_info, device,
+                                                   force=True)
                 except Exception:
                     LOG.exception(_LE('Failed to disconnect volume '
                                       '%(volume_id)s.'),
                                   {'volume_id': volume.id})
 
-            try:
-                volume.terminate_connection(volume, properties)
-            except Exception:
-                LOG.exception(_LE('Failed to terminate connection of volume '
-                                  '%(volume_id)s.'), {'volume_id': volume.id})
-
-            try:
-                client.volumes.detach(volume)
-            except Exception:
-                LOG.exception(_LE('Failed to detach volume %(volume_id)s.'),
-                              {'volume_id': volume.id})
+            if not volume.multiattach:
+                self.volume_api.attachment_delete(client, attachment.id)
 
     def _cinder_volume_data_iterator(self, client, volume, max_size, offset=0,
                                      chunk_size=None, partial_length=None):
@@ -760,7 +876,7 @@ class Store(glance_store.driver.Store):
         loc = location.store_location
         self._check_context(context)
         try:
-            client = self.get_cinderclient(context)
+            client = self.get_cinderclient(context, version='3.54')
             volume = client.volumes.get(loc.volume_id)
             size = int(volume.metadata.get('image_size',
                                            volume.size * units.Gi))
@@ -828,15 +944,15 @@ class Store(glance_store.driver.Store):
         """
 
         self._check_context(context, require_tenant=True)
-        client = self.get_cinderclient(context)
-        os_hash_value = hashlib.new(str(hashing_algo))
-        checksum = hashlib.md5()
+        client = self.get_cinderclient(context, version='3.54')
+        os_hash_value = utils.get_hasher(hashing_algo, False)
+        checksum = utils.get_hasher('md5', False)
         bytes_written = 0
         size_gb = int(math.ceil(float(image_size) / units.Gi))
         if size_gb == 0:
             size_gb = 1
         name = "image-%s" % image_id
-        owner = context.tenant
+        owner = context.project_id
         metadata = {'glance_image_id': image_id,
                     'image_size': str(image_size),
                     'image_owner': owner}
@@ -849,8 +965,18 @@ class Store(glance_store.driver.Store):
             LOG.info(_LI("Since image size is zero, we will be doing "
                          "resize-before-write for each GB which "
                          "will be considerably slower than normal."))
-        volume = client.volumes.create(size_gb, name=name, metadata=metadata,
-                                       volume_type=volume_type)
+        try:
+            volume = self.volume_api.create(client, size_gb, name=name,
+                                            metadata=metadata,
+                                            volume_type=volume_type)
+        except cinder_exception.NotFound:
+            LOG.error(_LE("Invalid volume type %s configured. Please check "
+                          "the `cinder_volume_type` configuration parameter."
+                          % volume_type))
+            msg = (_("Failed to create image-volume due to invalid "
+                     "`cinder_volume_type` configured."))
+            raise exceptions.BackendException(msg)
+
         volume = self._wait_volume_status(volume, 'creating', 'available')
         size_gb = volume.size
 
@@ -860,6 +986,10 @@ class Store(glance_store.driver.Store):
         try:
             while need_extend:
                 with self._open_cinder_volume(client, volume, 'wb') as f:
+                    # Sometimes the extended LUN on storage side takes time
+                    # to reflect in the device so we wait until the device
+                    # size is equal to the extended volume size.
+                    Store._wait_resize_device(volume, f)
                     f.seek(bytes_written)
                     if buf:
                         f.write(buf)
@@ -951,9 +1081,9 @@ class Store(glance_store.driver.Store):
         """
         loc = location.store_location
         self._check_context(context)
+        client = self.get_cinderclient(context)
         try:
-            volume = self.get_cinderclient(context).volumes.get(loc.volume_id)
-            volume.delete()
+            self.volume_api.delete(client, loc.volume_id)
         except cinder_exception.NotFound:
             raise exceptions.NotFound(image=loc.volume_id)
         except cinder_exception.ClientException as e:
