@@ -15,7 +15,8 @@
 
 import contextlib
 import errno
-import hashlib
+import io
+import math
 import os
 from unittest import mock
 
@@ -30,9 +31,12 @@ import fixtures
 from os_brick.initiator import connector
 from oslo_concurrency import processutils
 from oslo_config import cfg
+from oslo_utils.secretutils import md5
 from oslo_utils import units
 
 import glance_store as store
+from glance_store.common import attachment_state_manager
+from glance_store.common import cinder_utils
 from glance_store import exceptions
 from glance_store import location
 from glance_store.tests import base
@@ -87,25 +91,26 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
         fake_sc = [{u'endpoints': [{u'publicURL': u'http://foo/public_url'}],
                     u'endpoints_links': [],
                     u'name': u'cinder',
-                    u'type': u'volumev2'}]
+                    u'type': u'volumev3'}]
         self.context = FakeObject(service_catalog=fake_sc,
-                                  user='fake_user',
+                                  user_id='fake_user',
                                   auth_token='fake_token',
-                                  tenant='fake_tenant')
+                                  project_id='fake_project')
         self.fake_admin_context = mock.MagicMock()
         self.fake_admin_context.elevated.return_value = FakeObject(
             service_catalog=fake_sc,
-            user='admin_user',
+            user_id='admin_user',
             auth_token='admin_token',
-            tenant='admin_project')
+            project_id='admin_project')
+        cinder._reset_cinder_session()
 
     def test_location_url_prefix_is_set(self):
         self.assertEqual("cinder://cinder1", self.store.url_prefix)
 
     def test_get_cinderclient(self):
         cc = self.store.get_cinderclient(self.context)
-        self.assertEqual('fake_token', cc.client.auth_token)
-        self.assertEqual('http://foo/public_url', cc.client.management_url)
+        self.assertEqual('fake_token', cc.client.auth.token)
+        self.assertEqual('http://foo/public_url', cc.client.auth.endpoint)
 
     def test_get_cinderclient_with_user_overriden(self):
         self.config(cinder_store_user_name='test_user', group="cinder1")
@@ -113,16 +118,14 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
         self.config(cinder_store_project_name='test_project', group="cinder1")
         self.config(cinder_store_auth_address='test_address', group="cinder1")
         cc = self.store.get_cinderclient(self.context)
-        self.assertIsNone(cc.client.auth_token)
-        self.assertEqual('test_address', cc.client.management_url)
+        self.assertEqual('Default', cc.client.session.auth.project_domain_name)
+        self.assertEqual('test_project', cc.client.session.auth.project_name)
 
     def test_get_cinderclient_legacy_update(self):
         cc = self.store.get_cinderclient(self.fake_admin_context,
                                          legacy_update=True)
-        self.assertEqual('admin_token', cc.client.auth_token)
-        self.assertEqual('admin_user', cc.client.user)
-        self.assertEqual('admin_project', cc.client.projectid)
-        self.assertEqual('http://foo/public_url', cc.client.management_url)
+        self.assertEqual('admin_token', cc.client.auth.token)
+        self.assertEqual('http://foo/public_url', cc.client.auth.endpoint)
 
     def test_temporary_chown(self):
         class fake_stat(object):
@@ -183,11 +186,23 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
 
     def _test_open_cinder_volume(self, open_mode, attach_mode, error,
                                  multipath_supported=False,
-                                 enforce_multipath=False):
+                                 enforce_multipath=False,
+                                 encrypted_nfs=False, qcow2_vol=False,
+                                 multiattach=False):
         self.config(cinder_mount_point_base=None, group='cinder1')
-        fake_volume = mock.MagicMock(id=str(uuid.uuid4()), status='available')
-        fake_volumes = FakeObject(get=lambda id: fake_volume,
-                                  detach=mock.Mock())
+        fake_volume = mock.MagicMock(id=str(uuid.uuid4()), status='available',
+                                     multiattach=multiattach)
+        fake_volume.manager.get.return_value = fake_volume
+        fake_attachment_id = str(uuid.uuid4())
+        fake_attachment_create = {'id': fake_attachment_id}
+        if encrypted_nfs or qcow2_vol:
+            fake_attachment_update = mock.MagicMock(
+                id=fake_attachment_id,
+                connection_info={'driver_volume_type': 'nfs'})
+        else:
+            fake_attachment_update = mock.MagicMock(id=fake_attachment_id)
+        fake_conn_info = mock.MagicMock(connector={})
+        fake_volumes = FakeObject(get=lambda id: fake_volume)
         fake_client = FakeObject(volumes=fake_volumes)
         _, fake_dev_path = tempfile.mkstemp(dir=self.test_dir)
         fake_devinfo = {'path': fake_dev_path}
@@ -200,14 +215,22 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
             yield
 
         def do_open():
-            with self.store._open_cinder_volume(
-                    fake_client, fake_volume, open_mode):
-                if error:
-                    raise error
+            if multiattach:
+                with mock.patch.object(
+                        attachment_state_manager._AttachmentStateManager,
+                        'get_state') as mock_get_state:
+                    mock_get_state.return_value.__enter__.return_value = (
+                        attachment_state_manager._AttachmentState())
+                    with self.store._open_cinder_volume(
+                            fake_client, fake_volume, open_mode):
+                        pass
+            else:
+                with self.store._open_cinder_volume(
+                        fake_client, fake_volume, open_mode):
+                    if error:
+                        raise error
 
         def fake_factory(protocol, root_helper, **kwargs):
-            self.assertEqual(fake_volume.initialize_connection.return_value,
-                             kwargs['conn'])
             return fake_connector
 
         root_helper = "sudo glance-rootwrap /etc/glance/rootwrap.conf"
@@ -219,25 +242,92 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
                 mock.patch.object(cinder.Store, 'get_root_helper',
                                   return_value=root_helper), \
                 mock.patch.object(connector.InitiatorConnector, 'factory',
-                                  side_effect=fake_factory):
+                                  side_effect=fake_factory
+                                  ) as fake_conn_obj, \
+                mock.patch.object(cinder_utils.API,
+                                  'attachment_create',
+                                  return_value=fake_attachment_create
+                                  ) as attach_create, \
+                mock.patch.object(cinder_utils.API,
+                                  'attachment_update',
+                                  return_value=fake_attachment_update
+                                  ) as attach_update, \
+                mock.patch.object(cinder_utils.API,
+                                  'attachment_delete') as attach_delete, \
+                mock.patch.object(cinder_utils.API,
+                                  'attachment_get') as attach_get, \
+                mock.patch.object(cinder_utils.API,
+                                  'attachment_complete') as attach_complete, \
+                mock.patch.object(socket,
+                                  'gethostname') as mock_get_host, \
+                mock.patch.object(socket,
+                                  'getaddrinfo') as mock_get_host_ip:
+
+            fake_host = 'fake_host'
+            fake_addr_info = [[0, 1, 2, 3, ['127.0.0.1']]]
+            fake_ip = fake_addr_info[0][4][0]
+            mock_get_host.return_value = fake_host
+            mock_get_host_ip.return_value = fake_addr_info
 
             with mock.patch.object(connector,
-                                   'get_connector_properties') as mock_conn:
+                                   'get_connector_properties',
+                                   return_value=fake_conn_info) as mock_conn:
                 if error:
                     self.assertRaises(error, do_open)
+                elif encrypted_nfs or qcow2_vol:
+                    fake_volume.encrypted = False
+                    if encrypted_nfs:
+                        fake_volume.encrypted = True
+                    elif qcow2_vol:
+                        attach_get.return_value = mock.MagicMock(
+                            connection_info={'format': 'qcow2'})
+                    try:
+                        with self.store._open_cinder_volume(
+                                fake_client, fake_volume, open_mode):
+                            pass
+                    except exceptions.BackendException:
+                        attach_delete.assert_called_once_with(
+                            fake_client, fake_attachment_id)
                 else:
                     do_open()
-
-                mock_conn.assert_called_once_with(
-                    root_helper, socket.gethostname(), multipath_supported,
-                    enforce_multipath)
-                fake_connector.connect_volume.assert_called_once_with(mock.ANY)
-                fake_connector.disconnect_volume.assert_called_once_with(
-                    mock.ANY, fake_devinfo)
-                fake_volume.attach.assert_called_once_with(
-                    None, 'glance_store', attach_mode,
-                    host_name=socket.gethostname())
-                fake_volumes.detach.assert_called_once_with(fake_volume)
+                if not (encrypted_nfs or qcow2_vol):
+                    mock_conn.assert_called_once_with(
+                        root_helper, fake_ip,
+                        multipath_supported, enforce_multipath,
+                        host=fake_host)
+                    fake_connector.connect_volume.assert_called_once_with(
+                        mock.ANY)
+                    fake_connector.disconnect_volume.assert_called_once_with(
+                        mock.ANY, fake_devinfo, force=True)
+                    fake_conn_obj.assert_called_once_with(
+                        mock.ANY, root_helper, conn=mock.ANY,
+                        use_multipath=multipath_supported)
+                    attach_create.assert_called_once_with(
+                        fake_client, fake_volume.id, mode=attach_mode)
+                    attach_update.assert_called_once_with(
+                        fake_client, fake_attachment_id,
+                        fake_conn_info, mountpoint='glance_store')
+                    attach_complete.assert_called_once_with(
+                        fake_client, fake_attachment_id)
+                    attach_delete.assert_called_once_with(fake_client,
+                                                          fake_attachment_id)
+                else:
+                    mock_conn.assert_called_once_with(
+                        root_helper, fake_ip,
+                        multipath_supported, enforce_multipath,
+                        host=fake_host)
+                    fake_connector.connect_volume.assert_not_called()
+                    fake_connector.disconnect_volume.assert_not_called()
+                    fake_conn_obj.assert_called_once_with(
+                        mock.ANY, root_helper, conn=mock.ANY,
+                        use_multipath=multipath_supported)
+                    attach_create.assert_called_once_with(
+                        fake_client, fake_volume.id, mode=attach_mode)
+                    attach_update.assert_called_once_with(
+                        fake_client, fake_attachment_id,
+                        fake_conn_info, mountpoint='glance_store')
+                    attach_delete.assert_called_once_with(
+                        fake_client, fake_attachment_id)
 
     def test_open_cinder_volume_rw(self):
         self._test_open_cinder_volume('wb', 'rw', None)
@@ -261,6 +351,15 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
                                       multipath_supported=True,
                                       enforce_multipath=True)
 
+    def test_open_cinder_volume_nfs_encrypted(self):
+        self._test_open_cinder_volume('rb', 'ro', None, encrypted_nfs=True)
+
+    def test_open_cinder_volume_nfs_qcow2_volume(self):
+        self._test_open_cinder_volume('rb', 'ro', None, qcow2_vol=True)
+
+    def test_open_cinder_volume_multiattach_volume(self):
+        self._test_open_cinder_volume('rb', 'ro', None, multiattach=True)
+
     def test_cinder_check_context(self):
         self.assertRaises(exceptions.BadStoreConfiguration,
                           self.store._check_context, None)
@@ -271,27 +370,55 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
 
         self.store._check_context(FakeObject(service_catalog='fake'))
 
-    def test_cinder_configure_add(self):
+    def test_configure_add(self):
+
+        def fake_volume_type_check(name):
+            if name != 'some_type':
+                raise cinder.cinder_exception.NotFound(code=404)
+
         with mock.patch.object(self.store, 'get_cinderclient') as mocked_cc:
-            def raise_(ex):
-                raise ex
             mocked_cc.return_value = FakeObject(volume_types=FakeObject(
-                find=lambda name: 'some_type' if name == 'some_type'
-                else raise_(cinder.cinder_exception.NotFound(code=404))))
+                find=fake_volume_type_check))
             self.config(cinder_volume_type='some_type',
                         group=self.store.backend_group)
             # If volume type exists, no exception is raised
             self.store.configure_add()
-            # setting cinder_volume_type to non-existent value will raise
-            # BadStoreConfiguration exception
+            # setting cinder_volume_type to non-existent value will log a
+            # warning
             self.config(cinder_volume_type='some_random_type',
                         group=self.store.backend_group)
+            with mock.patch.object(cinder, 'LOG') as mock_log:
+                self.store.configure_add()
+                mock_log.warning.assert_called_with(
+                    "Invalid `cinder_volume_type some_random_type`")
 
-            self.assertRaises(exceptions.BadStoreConfiguration,
-                              self.store.configure_add)
-            # when only 1 store is configured, BackendException is raised
-            self.config(enabled_backends={'cinder1': 'cinder'})
-            self.assertRaises(exceptions.BackendException,
+    def test_configure_add_cinder_service_down(self):
+
+        def fake_volume_type_check(name):
+            raise cinder.cinder_exception.ClientException(code=503)
+
+        self.config(cinder_volume_type='some_type',
+                    group=self.store.backend_group)
+        with mock.patch.object(self.store, 'get_cinderclient') as mocked_cc:
+            mocked_cc.return_value = FakeObject(volume_types=FakeObject(
+                find=fake_volume_type_check))
+            # We handle the ClientException to pass so no exception is raised
+            # in this case
+            self.store.configure_add()
+
+    def test_configure_add_authorization_failed(self):
+
+        def fake_volume_type_check(name):
+            raise cinder.exceptions.AuthorizationFailure(code=401)
+
+        self.config(cinder_volume_type='some_type',
+                    group=self.store.backend_group)
+        with mock.patch.object(self.store, 'get_cinderclient') as mocked_cc:
+            mocked_cc.return_value = FakeObject(volume_types=FakeObject(
+                find=fake_volume_type_check))
+            # Anything apart from invalid volume type or cinder service
+            # down will raise an exception
+            self.assertRaises(cinder.exceptions.AuthorizationFailure,
                               self.store.configure_add)
 
     def test_is_image_associated_with_store(self):
@@ -398,12 +525,13 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
             self.assertEqual(expected_image_size, image_size)
 
     def _test_cinder_add(self, fake_volume, volume_file, size_kb=5,
-                         verifier=None, backend="cinder1"):
+                         verifier=None, backend="cinder1", fail_resize=False):
         expected_image_id = str(uuid.uuid4())
         expected_size = size_kb * units.Ki
         expected_file_contents = b"*" * expected_size
         image_file = six.BytesIO(expected_file_contents)
-        expected_checksum = hashlib.md5(expected_file_contents).hexdigest()
+        expected_checksum = md5(expected_file_contents,
+                                usedforsecurity=False).hexdigest()
         expected_location = 'cinder://%s/%s' % (backend, fake_volume.id)
         fake_client = FakeObject(auth_token=None, management_url=None)
         fake_volume.manager.get.return_value = fake_volume
@@ -417,7 +545,11 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
 
         with mock.patch.object(cinder.Store, 'get_cinderclient') as mock_cc, \
                 mock.patch.object(self.store, '_open_cinder_volume',
-                                  side_effect=fake_open):
+                                  side_effect=fake_open), \
+                mock.patch.object(
+                    cinder.Store, '_wait_resize_device') as mock_wait_resize:
+            if fail_resize:
+                mock_wait_resize.side_effect = exceptions.BackendException()
             mock_cc.return_value = FakeObject(client=fake_client,
                                               volumes=fake_volumes)
             loc, size, checksum, metadata = self.store.add(expected_image_id,
@@ -431,7 +563,7 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
             fake_volumes.create.assert_called_once_with(
                 1,
                 name='image-%s' % expected_image_id,
-                metadata={'image_owner': self.context.tenant,
+                metadata={'image_owner': self.context.project_id,
                           'glance_image_id': expected_image_id,
                           'image_size': str(expected_size)},
                 volume_type='some_type')
@@ -468,8 +600,7 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
     def test_cinder_delete(self):
         fake_client = FakeObject(auth_token=None, management_url=None)
         fake_volume_uuid = str(uuid.uuid4())
-        fake_volume = FakeObject(delete=mock.Mock())
-        fake_volumes = {fake_volume_uuid: fake_volume}
+        fake_volumes = FakeObject(delete=mock.Mock())
 
         with mock.patch.object(cinder.Store, 'get_cinderclient') as mocked_cc:
             mocked_cc.return_value = FakeObject(client=fake_client,
@@ -480,7 +611,7 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
                                                              "cinder1",
                                                              conf=self.conf)
             self.store.delete(loc, context=self.context)
-            fake_volume.delete.assert_called_once_with()
+            fake_volumes.delete.assert_called_once_with(fake_volume_uuid)
 
     def test_cinder_add_different_backend(self):
         self.store = cinder.Store(self.conf, backend="cinder2")
@@ -492,3 +623,43 @@ class TestMultiCinderStore(base.MultiStoreBaseTest,
                                      size=1)
         volume_file = six.BytesIO()
         self._test_cinder_add(fake_volume, volume_file, backend="cinder2")
+
+    def test__get_device_size(self):
+        fake_data = b"fake binary data"
+        fake_len = int(math.ceil(float(len(fake_data)) / units.Gi))
+        fake_file = io.BytesIO(fake_data)
+        dev_size = cinder.Store._get_device_size(fake_file)
+        self.assertEqual(fake_len, dev_size)
+
+    @mock.patch.object(time, 'sleep')
+    def test__wait_resize_device_resized(self, mock_sleep):
+        fake_vol = mock.MagicMock()
+        fake_vol.size = 2
+        fake_file = io.BytesIO(b"fake binary data")
+        with mock.patch.object(
+                cinder.Store, '_get_device_size') as mock_get_dev_size:
+            mock_get_dev_size.side_effect = [1, 2]
+            cinder.Store._wait_resize_device(fake_vol, fake_file)
+
+    @mock.patch.object(time, 'sleep')
+    def test__wait_resize_device_fails(self, mock_sleep):
+        fake_vol = mock.MagicMock()
+        fake_vol.size = 2
+        fake_file = io.BytesIO(b"fake binary data")
+        with mock.patch.object(
+                cinder.Store, '_get_device_size',
+                return_value=1):
+            self.assertRaises(
+                exceptions.BackendException,
+                cinder.Store._wait_resize_device,
+                fake_vol, fake_file)
+
+    def test_cinder_add_fail_resize(self):
+        volume_file = io.BytesIO()
+        fake_volume = mock.MagicMock(id=str(uuid.uuid4()),
+                                     status='available',
+                                     size=1)
+        self.assertRaises(exceptions.BackendException,
+                          self._test_cinder_add, fake_volume, volume_file,
+                          fail_resize=True)
+        fake_volume.delete.assert_called_once()
